;*              Copyright 1992 Digital Equipment Corporation
;*                         All Rights Reserved
;*
;* Permission to use, copy, and modify this software and its documentation is
;* hereby granted only under the following terms and conditions.  Both the
;* above copyright notice and this permission notice must appear in all copies
;* of the software, derivative works or modified versions, and any portions
;* thereof, and both notices must appear in supporting documentation.
;*
;* Users of this software agree to the terms and conditions set forth herein,
;* and hereby grant back to Digital a non-exclusive, unrestricted, royalty-free
;* right and license under any changes, enhancements or extensions made to the
;* core functions of the software, including but not limited to those affording
;* compatibility with other hardware or software environments, but excluding
;* applications which incorporate this software.  Users further agree to use
;* their best efforts to return to Digital any such changes, enhancements or
;* extensions that they make and inform Digital of noteworthy uses of this
;* software.  Correspondence should be provided to Digital at:
;* 
;*			Director, Cambridge Research Lab
;*			Digital Equipment Corp
;*			One Kendall Square, Bldg 700
;*			Cambridge MA 02139
;* 
;* This software may be distributed (but not offered for sale or transferred
;* for compensation) to third parties, provided such third parties agree to
;* abide by the terms and conditions of this notice.  
;* 
;* THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
;* WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
;* MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
;* CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
;* DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
;* PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
;* ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
;* SOFTWARE.

; $Id: runtime.scm,v 1.22 1992/09/18 23:56:44 birkholz Exp $

(for-each dylan::load
	  '("class"			; Dylan Class data type.
	    "generic"			; Generic dispatch mechanism.
	    "class-structure"		; The heterarchy.
	    "runtime-top"		; Common parameter lists, etc.
	    "runtime-internal"		; Procedures called by
					; expansion of code, not by
					; users
	    "runtime-methods"		; Predefined methods
	    "runtime-functions"		; Predefined functions and
					; generic functions
	    "runtime-bitstrings"	; Integers as bitstrings
	    "runtime-collections"	; Collections, sequences, ...
	    "runtime-collections-iterate" ; Iteration + Collection Keys
	    "runtime-collections-generic1" ; Collection generic functions 
	    "runtime-collections-generic2" ; More generic collection operations.
	    "runtime-collections-array"	; Arrays
	    "runtime-collections-deque"	; Deques
	    "runtime-collections-list"	; Lists
	    "runtime-collections-range"	; Ranges
	    "runtime-collections-string" ; Strings
	    "runtime-collections-table"	; Tables
	    "runtime-collections-vector" ; Vectors
	    "runtime-exceptions"))	; Exceptions
